"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.HasuraClient = void 0;
const apollo_boost_1 = require("apollo-boost");
const apollo_link_http_1 = require("apollo-link-http");
const util_1 = __importStar(require("@cardano-graphql/util"));
const child_process_1 = require("child_process");
const cross_fetch_1 = __importDefault(require("cross-fetch"));
const graphql_1 = require("graphql");
const wrap_1 = require("@graphql-tools/wrap");
const p_retry_1 = __importDefault(require("p-retry"));
const path_1 = __importDefault(require("path"));
const ts_log_1 = require("ts-log");
const bignumber_js_1 = __importDefault(require("bignumber.js"));
class HasuraClient {
    constructor(hasuraCliPath, hasuraUri, pollingInterval, logger = ts_log_1.dummyLogger) {
        this.hasuraCliPath = hasuraCliPath;
        this.hasuraUri = hasuraUri;
        this.logger = logger;
        this.applyingSchemaAndMetadata = false;
        this.adaCirculatingSupplyFetcher = new util_1.DataFetcher('AdaCirculatingSupply', this.getAdaCirculatingSupply.bind(this), pollingInterval, this.logger);
        this.client = new apollo_boost_1.ApolloClient({
            cache: new apollo_boost_1.InMemoryCache({
                addTypename: false
            }),
            defaultOptions: {
                query: {
                    fetchPolicy: 'network-only'
                }
            },
            link: apollo_link_http_1.createHttpLink({
                uri: `${this.hasuraUri}/v1/graphql`,
                fetch: cross_fetch_1.default,
                headers: {
                    'X-Hasura-Role': 'cardano-graphql'
                }
            })
        });
    }
    getAdaCirculatingSupply() {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.client.query({
                query: apollo_boost_1.gql `query {
          utxos_aggregate {
              aggregate {
                  sum {
                      value
                  }
              }
          }
      }`
            });
            return result.data.utxos_aggregate.aggregate.sum.value;
        });
    }
    hasuraCli(command) {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                child_process_1.exec(`${this.hasuraCliPath} --skip-update-check --project ${path_1.default.resolve(__dirname, '..', 'hasura', 'project')} --endpoint ${this.hasuraUri} ${command}`, (error, stdout) => {
                    if (error) {
                        reject(error);
                    }
                    this.logger.debug(stdout);
                    resolve();
                });
            });
        });
    }
    initialize() {
        return __awaiter(this, void 0, void 0, function* () {
            this.logger.info('Initializing Hasura', { module: 'HasuraClient' });
            yield this.applySchemaAndMetadata();
            yield p_retry_1.default(() => __awaiter(this, void 0, void 0, function* () {
                this.schema = yield this.buildHasuraSchema();
            }), {
                factor: 1.75,
                retries: 9,
                onFailedAttempt: util_1.default.onFailedAttemptFor('Fetching Hasura schema via introspection', this.logger)
            });
            this.logger.info('Hasura initialized', { module: 'HasuraClient' });
            yield this.adaCirculatingSupplyFetcher.initialize();
        });
    }
    shutdown() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.adaCirculatingSupplyFetcher.shutdown();
        });
    }
    applySchemaAndMetadata() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.applyingSchemaAndMetadata)
                return;
            this.applyingSchemaAndMetadata = true;
            yield p_retry_1.default(() => __awaiter(this, void 0, void 0, function* () {
                yield this.hasuraCli('migrate apply --down all');
                yield this.hasuraCli('migrate apply --up all');
            }), {
                factor: 1.75,
                retries: 9,
                onFailedAttempt: util_1.default.onFailedAttemptFor('Applying PostgreSQL schema migrations', this.logger)
            });
            yield p_retry_1.default(() => __awaiter(this, void 0, void 0, function* () {
                yield this.hasuraCli('metadata clear');
                yield this.hasuraCli('metadata apply');
            }), {
                factor: 1.75,
                retries: 9,
                onFailedAttempt: util_1.default.onFailedAttemptFor('Applying Hasura metadata', this.logger)
            });
            this.applyingSchemaAndMetadata = false;
        });
    }
    buildHasuraSchema() {
        return __awaiter(this, void 0, void 0, function* () {
            const executor = ({ document, variables }) => __awaiter(this, void 0, void 0, function* () {
                const query = graphql_1.print(document);
                try {
                    const fetchResult = yield cross_fetch_1.default(`${this.hasuraUri}/v1/graphql`, {
                        method: 'POST',
                        headers: {
                            'Content-Type': 'application/json',
                            'X-Hasura-Role': 'cardano-graphql'
                        },
                        body: JSON.stringify({ query, variables })
                    });
                    return fetchResult.json();
                }
                catch (error) {
                    this.logger.error(error);
                    throw error;
                }
            });
            const coreTypes = [
                'Block',
                'Cardano',
                'Epoch',
                'Block',
                'Transaction'
            ];
            const schema = wrap_1.wrapSchema({
                schema: yield wrap_1.introspectSchema(executor),
                executor
            });
            for (const t of coreTypes) {
                const gqlType = schema.getType(t);
                if (!gqlType) {
                    throw new Error(`Remote schema is missing ${t}`);
                }
            }
            return schema;
        });
    }
    getPaymentAddressSummary(address, atBlock) {
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.client.query({
                query: apollo_boost_1.gql `query PaymentAddressSummary (
          $address: String!
          $atBlock: Int
      ){
          utxos (
              where: {
                  _and: {
                      address: { _eq: $address },
                      transaction: { block: { number: { _lte: $atBlock }}}
                  }
              }
          ) {
              value
              tokens {
                  assetId
                  assetName
                  policyId
                  quantity
              }
          }
          utxos_aggregate (
              where: {
                  _and: {
                      address: { _eq: $address },
                      transaction: { block: { number: { _lte: $atBlock }}}
                  }
              }
          ) {
              aggregate {
                  count
              }
          }
      }`,
                variables: { address, atBlock }
            });
            const map = new Map();
            for (const utxo of result.data.utxos) {
                if (map.has('ada')) {
                    const current = map.get('ada');
                    map.set('ada', Object.assign(Object.assign({}, current), {
                        quantity: new bignumber_js_1.default(current.quantity)
                            .plus(new bignumber_js_1.default(utxo.value))
                            .toString()
                    }));
                }
                else {
                    map.set('ada', {
                        assetId: 'ada',
                        assetName: 'ada',
                        policyId: '',
                        quantity: utxo.value
                    });
                }
                for (const token of utxo.tokens) {
                    if (map.has(token.assetId)) {
                        const current = map.get(token.assetId);
                        map.set(token.assetId, Object.assign(Object.assign({}, current), {
                            quantity: new bignumber_js_1.default(current.quantity)
                                .plus(new bignumber_js_1.default(token.quantity))
                                .toString()
                        }));
                    }
                    else {
                        map.set(token.assetId, token);
                    }
                }
            }
            return {
                assetBalances: [...map.values()],
                utxosCount: result.data.utxos_aggregate.aggregate.count
            };
        });
    }
    getMeta(nodeTipBlockNumber) {
        var _a, _b, _c;
        return __awaiter(this, void 0, void 0, function* () {
            const result = yield this.client.query({
                query: apollo_boost_1.gql `query {
          epochs (limit: 1, order_by: { number: desc }) {
              number
          }
          cardano {
              tip {
                  epoch {
                      number
                  }
                  number
                  forgedAt
              }
          }}`
            });
            const { tip } = (_a = result.data) === null || _a === void 0 ? void 0 : _a.cardano[0];
            const lastEpoch = (_b = result.data) === null || _b === void 0 ? void 0 : _b.epochs[0];
            return {
                initialized: lastEpoch.number === ((_c = tip.epoch) === null || _c === void 0 ? void 0 : _c.number),
                syncPercentage: (tip.number / nodeTipBlockNumber) * 100
            };
        });
    }
}
exports.HasuraClient = HasuraClient;
//# sourceMappingURL=HasuraClient.js.map